import * as i0 from '@angular/core';
import { Directive, Input, NgModule } from '@angular/core';
import { coerceBooleanProperty } from '@angular/cdk/coercion';
import { createListTemplateManager, RxDefaultListViewContext } from '@rx-angular/cdk/template';
import { coerceDistinctWith } from '@rx-angular/cdk/coercing';
import { ReplaySubject, Subscription } from 'rxjs';
import * as i1 from '@rx-angular/cdk/render-strategies';

/**
 * @Directive RxFor
 *
 * @description
 *
 * The `*rxFor` structural directive provides a convenient and performant way for rendering
 * templates out of a list of items.
 * Input values can be provided either as `Observable`, `Promise` or `static` values. Just as the `*ngFor` directive, the
 * `*rxFor` is placed on an
 * element, which becomes the parent of the cloned templates.
 *
 * The `RxFor` implements `EmbeddedView Rendering`.
 * Compared to the `NgForOf`, `RxFor` treats each child template as single renderable entity. For each
 * change in the provided list of items it will apply and detect changes to only affected views.
 *
 * Under the hood, it leverages the power of the `StrategyCredential`s which in turn take care of scheduling and
 * prioritizing the change detection for each child template (aka item in the list).
 * This way the rendering behavior of each instance of `RxFor` can be configured individually.
 *
 * `RxStrategyCredentials` and `EmbeddedView Rendering` together build the basis for the `concurrent mode`. Based on
 * the configured strategy every template will get processed in an individual task, which enables chunked and
 * cancellable rendering of the list.
 *
 * As further improvement compared to the basic `*ngFor` implementation, `*rxFor` is able to take care of
 * `ChangeDetection` in situations which include `projected views` (aka `@ContentChild` or `@ViewChild`).
 * Learn more about this in the example section.
 *
 *
 * ### Context Variables
 *
 * The following context variables are available for each template:
 *
 * - $implicit: `T` // the default variable accessed by `let val`
 * - item$: `Observable<T>` // the same value as $implicit, but as `Observable`
 * - index: `number` // current index of the item
 * - count: `number` // count of all items in the list
 * - first: `boolean` // true if the item is the first in the list
 * - last: `boolean` // true if the item is the last in the list
 * - even: `boolean` // true if the item has on even index (index % 2 === 0)
 * - odd: `boolean` // the opposite of even
 * - index$: `Observable<number>` // index as `Observable`
 * - count$: `Observable<number>` // count as `Observable`
 * - first$: `Observable<boolean>` // first as `Observable`
 * - last$: `Observable<boolean>` // last as `Observable`
 * - even$: `Observable<boolean>` // even as `Observable`
 * - odd$: `Observable<boolean>` // odd as `Observable`
 * - select: `(keys: (keyof T)[], distinctByMap) => Observable<Partial<T>>` // returns a selection function which
 * accepts an array of properties to pluck out of every list item. The function returns the selected properties of
 * the current list item as distinct `Observable` key-value-pair. See the example below:
 *
 * This example showcases the `select` view-context function used for deeply nested lists.
 *
 *  ```html
 * <ul>
 *   <li *rxFor="let hero of heroes$; trackBy: trackItem; let select = select;">
 *     <div>
 *       <strong>{{ hero.name }}</strong></br>
 *       Defeated enemies:
 *     </div>
 *      <span *rxFor="let enemy of select(['defeatedEnemies']); trackBy: trackEnemy;">
 *        {{ enemy.name }}
 *      </span>
 *   </li>
 * </ul>
 *  ```
 *
 * ### Input properties
 *
 *  - trackBy: `(index: number, item: T) => any`
 *  - trackBy: `keyof T`
 *  - strategy: `string`
 *  - strategy: `Observable<string>`
 *  - parent: `boolean`;
 *  - renderCallback: `Subject<T[]>`
 *
 *
 * ### Features of `*rxFor`
 *
 * Included features for `*rxFor`:
 * - Push based architecture
 * - Immutable as well as mutable data structures (`trackBy`)
 * - Provide a comprehensive set of context variables for each view
 * - Provide a way to fix `ChangeDetection` issues in `Projected Views` scenarios
 * - automatically runs out of `NgZone`, provide an easy way to opt-in (`patchZone`)
 * - Notify about when rendering of child templates is finished (`renderCallback`)
 * - Reactive as well as imperative values in the template (`ngFor` drop-in replacement)
 * - `ListManager`: special logic for differ mechanism to avoid over-rendering; abstracts away low level logic
 * - render every `EmbeddedView` on its own while applying the configured `RxStrategyCredentials#behavior`
 * - cancel any scheduled work if a remove was triggered for a `trackById`
 * - cancel any update if a new update was triggered for the same `trackById`
 *
 *
 * ### Simple example using `*rxFor` with `Observable` values
 * ```html
 * <ul>
 *   <li *rxFor="let item of observableItems$; trackBy: trackItem">
 *      {{ item }}
 *   </li>
 * </ul>
 * ```
 *
 * ### Simple example using `*rxFor` with simple static values
 * ```html
 * <ul>
 *   <li *rxFor="let item of items; trackBy: trackItem">
 *      {{ item }}
 *   </li>
 * </ul>
 * ```
 *
 *
 * ### Using the context variables
 *
 * ```html
 * <ul>
 *   <li
 *     *rxFor="
 *       let item of observableItems$;
 *       let count = count;
 *       let index = index;
 *       let first = first;
 *       let last = last;
 *       let even = even;
 *       let odd = odd;
 *       trackBy: trackItem;
 *     "
 *   >
 *     <div>{{ count }}</div>
 *     <div>{{ index }}</div>
 *     <div>{{ item }}</div>
 *     <div>{{ first }}</div>
 *     <div>{{ last }}</div>
 *     <div>{{ even }}</div>
 *     <div>{{ odd }}</div>
 *   </li>
 * </ul>
 * ```
 *
 * ### Projected Views (`parent`)
 *
 * Imagine the following situation:
 *
 * ```ts
 * \@Component({
 *   selector: 'app-list-component',
 *   template: `
 *     <ng-content select="app-list-item"></ng-content>
 *   `
 * })
 * export class AppListComponent {
 *  \@ContentChildren(AppListItemComponent) appListItems: QueryList<AppListItemComponent>:
 * }
 * ```
 *
 * `AppListComponent` has a `contentOutlet` where it expects `AppListItemComponents` to be inserted into. In this case
 * `AppListComponent`s state is dependent on its `ContentChildren`.
 * This situation leads to the problem that `AppListComponent` needs to get informed about updates of its child views.
 * This is a known issue which has never been solved for `ngFor` (or other structural directives) especially in
 * combination with `CD OnPush` see here: (https://github.com/angular/angular/pull/35428)
 * `RxFor` solves this issue for you by providing a simple input parameter `parent: boolean`.
 * If set to `true`, `*rxFor` will automatically detect every other `Component` where its
 * `EmbeddedView`s were inserted into. Those components will get change detected as well in order to force
 * update their state accordingly.
 *
 * The usage of `AppListComponent` looks like this:
 *
 * ```html
 * <app-list-component>
 *   <app-list-item
 *     *rxFor="
 *       let item of observableItems$;
 *       parent: true;
 *     "
 *   >
 *     <div>{{ item }}</div>
 *   </app-list-item>
 * </app-list-component>
 * ```
 * ### `NgZone` patch
 *
 * By default `*rxFor` will create it's `EmbeddedViews` outside of `NgZone` which drastically speeds up the
 * performance.
 * There are scenarios where you want to opt-in to `NgZone` though. If views are created out of `NgZone`, all
 * `EventListeners` attached to them run out `NgZone` as well.
 *
 * Take a look at the following example:
 *
 * ```ts
 * \@Component({
 *   selector: 'app-root',
 *   template: `
 *     <!-- clickedHeroName won't get updated due to `NgZone` not noticing the click -->
 *     {{ clickedHeroName }}
 *     <ng-container *rxFor="let hero of heroes$; trackBy: trackHero">
 *       <!-- click runs out of `NgZone` -->
 *       <button (click)="heroClicked(hero)">{{ hero.name }}</button>
 *     </ng-container>
 *   `
 * })
 * export class AppComponent {
 *   clickedHeroName = '';
 *
 *   heroClicked(hero: Hero) {
 *     // this will run out of `NgZone` and thus not update the DOM
 *     this.clickedHeroName = hero.name;
 *   }
 * }
 * ```
 *
 * There are several ways to get around this issue.
 * `*rxFor` can be configured to create views inside of `NgZone` with the `patchZone` flag:
 *
 * ```html
 * <ng-container *rxFor="let hero of heroes$; trackBy: trackHero; patchZone: true">
 *   <!-- click now gets detected by `NgZone` -->
 *   <button (click)="heroClicked(hero)">{{ hero.name }}</button>
 * </ng-container>
 * ```
 *
 * However, `patchZone: true` can in some cases have a negative impact on the performance of the `*rxFor` Directive.
 * Since the creation of the `EmbeddedViews` will most likely happen in batches, every batch will result in one
 * `NgZone` cycle resulting in a possible re-rendering of many other `Components`.
 *
 * Another approach would be to manually detect changes coming from `unpatched` EventListeners or wrapping them in
 * `NgZone`.
 *
 * ```ts
 * export class AppComponent {
 *   clickedHeroName = '';
 *
 *   constructor(
 *     private cdRef: ChangeDetectorRef, // option1
 *     private ngZone: NgZone // option 2
 *   ) {}
 *
 *   heroClicked(hero: Hero) {
 *     // this will run out of `NgZone` and thus not update the DOM
 *     this.clickedHeroName = hero.name;
 *     this.cdRef.markForCheck(); // option 1
 *
 *     // option 2
 *     this.ngZone.run(() => this.clickedHeroName = hero.name);
 *   }
 * }
 * ```
 *
 *
 * @docsCategory RxFor
 * @docsPage RxFor
 * @publicApi
 */
/* @todo: rename to ForDirective? */
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class RxFor {
    /** @internal */
    constructor(iterableDiffers, cdRef, ngZone, eRef, templateRef, viewContainerRef, strategyProvider, errorHandler) {
        this.iterableDiffers = iterableDiffers;
        this.cdRef = cdRef;
        this.ngZone = ngZone;
        this.eRef = eRef;
        this.templateRef = templateRef;
        this.viewContainerRef = viewContainerRef;
        this.strategyProvider = strategyProvider;
        this.errorHandler = errorHandler;
        /* @todo: rename to `rxRenderParent`? */
        /* eslint-disable @angular-eslint/no-input-rename */
        /**
         * @description
         *  If `parent` is set to `true` (default to `false`), `*rxFor` will automatically detect every other `Component` where its
         * `EmbeddedView`s were inserted into. Those components will get change detected as well in order to force
         * update their state accordingly. In the given example, `AppListComponent` will get notified about which insert
         * or remove any `AppListItemComponent`.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *    <app-list-component>
         *      <app-list-item
         *        *rxFor="
         *          let item of items$;
         *          trackBy: trackItem;
         *          parent: true;
         *        "
         *      >
         *        <div>{{ item.name }}</div>
         *      </app-list-item>
         *    </app-list-component>
         *   `
         * })
         * export class AppComponent {
         *   items$ = itemService.getItems();
         * }
         *
         * @param renderParent
         */
        this.renderParent = true;
        /**
         * @description
         * A flag to control whether *rxFor templates are created within `NgZone` or not.
         * By default `*rxFor` will create it's `EmbeddedViews` outside of `NgZone` which drastically speeds up the
         * performance.
         * If `patchZone` is set to `true` (defaults to `false`), `*rxFor` will create its EmbeddedViews inside of `NgZone`.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *    <app-list-component>
         *      <app-list-item
         *        *rxFor="
         *          let item of items$;
         *          trackBy: trackItem;
         *          patchZone: true;
         *        "
         *      >
         *        <div>{{ item.name }}</div>
         *      </app-list-item>
         *    </app-list-component>
         *   `
         * })
         * export class AppComponent {
         *   items$ = itemService.getItems();
         * }
         *
         * @param patchZone
         */
        this.patchZone = this.strategyProvider.config.patchZone;
        /** @internal */
        this.strategyInput$ = new ReplaySubject(1);
        /** @internal */
        this.observables$ = new ReplaySubject(1);
        /** @internal */
        this.values$ = this.observables$.pipe(coerceDistinctWith());
        /** @internal */
        this.strategy$ = this.strategyInput$.pipe(coerceDistinctWith());
        /** @internal */
        this._subscription = Subscription.EMPTY;
        /** @internal */
        this._trackBy = (i, a) => a;
        /** @internal */
        this._distinctBy = (a, b) => a === b;
    }
    /**
     * @description
     * The iterable input
     *
     * @example
     * <ng-container *rxFor="heroes$; let hero">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     *
     * @param potentialObservable
     */
    set rxFor(potentialObservable) {
        this.observables$.next(potentialObservable);
    }
    /** @internal */
    set rxForOf(potentialObservable) {
        this.observables$.next(potentialObservable);
    }
    /**
     * @description
     * The rendering strategy to be used for each template of the list of items.
     * Use it to dynamically manage your rendering strategy. You can switch the strategies
     * imperatively (with a string) or by binding an Observable.
     * The default strategy is `'normal'`.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxFor="let hero of heroes$; strategy: strategy">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy = 'low';
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxFor="let hero of heroes$; strategy: strategy$">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy$ = new BehaviorSubject('immediate');
     * }
     *
     * @param strategyName
     * @see {@link strategies}
     */
    set strategy(strategyName) {
        this.strategyInput$.next(strategyName);
    }
    /**
     * @description
     * A function or key that defines how to track changes for items in the iterable.
     *
     * When items are added, moved, or removed in the iterable,
     * the directive must re-render the appropriate DOM nodes.
     * To minimize churn in the DOM, only nodes that have changed
     * are re-rendered.
     *
     * By default, rxFor assumes that the object instance identifies the node in the iterable (equality check `===`).
     * When a function or key is supplied, rxFor uses the result to identify the item node.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <app-list-component>
     *      <app-list-item
     *        *rxFor="
     *          let item of items$;
     *          trackBy: 'id';
     *        "
     *      >
     *        <div>{{ item.name }}</div>
     *      </app-list-item>
     *    </app-list-component>
     *   `
     * })
     * export class AppComponent {
     *   items$ = itemService.getItems();
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <app-list-component>
     *      <app-list-item
     *        *rxFor="
     *          let item of items$;
     *          trackBy: trackItem;
     *        "
     *      >
     *        <div>{{ item.name }}</div>
     *      </app-list-item>
     *    </app-list-component>
     *   `
     * })
     * export class AppComponent {
     *   items$ = itemService.getItems();
     *   trackItem = (idx, item) => item.id;
     * }
     *
     * @param trackByFnOrKey
     */
    set trackBy(trackByFnOrKey) {
        this._trackBy =
            typeof trackByFnOrKey !== 'function'
                ? (i, a) => a[trackByFnOrKey]
                : trackByFnOrKey;
    }
    /**
     * @internal
     * A function that defines how to track `updates` of items.
     * In addition to track when items are added, moved, or removed you can provide a function that determines if any
     * updates happened to an item. Use this is if you want to have even more control about what changes lead to
     * re-renderings of the DOM.
     *
     * By default, rxFor identifies if an update happens by doing an (equality check `===`).
     * When a function supplied, rxFor uses the result to identify the item node.
     *
     * @internal
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <app-list-component>
     *      <app-list-item
     *        *rxFor="
     *          let item of items$;
     *          trackBy: trackItem;
     *          distinctBy: distinctItem;
     *        "
     *      >
     *        <div>{{ item.name }}</div>
     *      </app-list-item>
     *    </app-list-component>
     *   `
     * })
     * export class AppComponent {
     *   items$ = itemService.getItems();
     *   trackItem = (idx, item) => item.id;
     *   // only changes to the name lead to a re-rendering of a child template
     *   distinctItem = (itemA, itemB) => itemA.name === itemB.name;
     * }
     *
     * @param distinctBy
     */
    /*@Input('rxForDistinctBy')*/
    set distinctBy(distinctBy) {
        this._distinctBy = distinctBy;
    }
    /**
     * @description
     * A `Subject` which emits whenever *rxFor finished rendering a set changes to the view.
     * This enables developers to perform actions when a list has finished rendering.
     * The `renderCallback` is useful in situations where you rely on specific DOM properties like the `height` a
     * table after all items got rendered.
     * It is also possible to use the renderCallback in order to determine if a view should be visible or not. This
     * way developers can hide a list as long as it has not finished rendering.
     *
     * The result of the `renderCallback` will contain the currently rendered set of items in the iterable.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *    <app-list-component>
     *      <app-list-item
     *        *rxFor="
     *          let item of items$;
     *          trackBy: trackItem;
     *          renderCallback: itemsRendered;
     *        "
     *      >
     *        <div>{{ item.name }}</div>
     *      </app-list-item>
     *    </app-list-component>
     *   `
     * })
     * export class AppComponent {
     *   items$: Observable<Item[]> = itemService.getItems();
     *   trackItem = (idx, item) => item.id;
     *   // only changes to the name lead to a re-rendering of a child template
     *   distinctItem = (itemA, itemB) => itemA.name === itemB.name;
     *   // this emits whenever rxFor finished rendering changes
     *   itemsRendered = new Subject<Item[]>();
     * }
     *
     * @param renderCallback
     */
    set renderCallback(renderCallback) {
        this._renderCallback = renderCallback;
    }
    /** @internal */
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
    /** @internal */
    ngOnInit() {
        this.listManager = createListTemplateManager({
            iterableDiffers: this.iterableDiffers,
            renderSettings: {
                cdRef: this.cdRef,
                eRef: this.eRef,
                strategies: this.strategyProvider.strategies,
                defaultStrategyName: this.strategyProvider.primaryStrategy,
                parent: coerceBooleanProperty(this.renderParent),
                patchZone: this.patchZone ? this.ngZone : false,
                errorHandler: this.errorHandler,
            },
            templateSettings: {
                viewContainerRef: this.viewContainerRef,
                templateRef: this.templateRef,
                createViewContext: this.createViewContext,
                updateViewContext: this.updateViewContext,
            },
            trackBy: this._trackBy,
        });
        this.listManager.nextStrategy(this.strategy$);
        this._subscription = this.listManager
            .render(this.values$)
            .subscribe((v) => { var _a; return (_a = this._renderCallback) === null || _a === void 0 ? void 0 : _a.next(v); });
    }
    /** @internal */
    createViewContext(item, computedContext) {
        return new RxDefaultListViewContext(item, computedContext);
    }
    /** @internal */
    updateViewContext(item, view, computedContext) {
        view.context.updateContext(computedContext);
        view.context.$implicit = item;
    }
    /** @internal */
    ngOnDestroy() {
        this._subscription.unsubscribe();
        this.viewContainerRef.clear();
    }
}
/** @nocollapse */ /** @nocollapse */ RxFor.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: RxFor, deps: [{ token: i0.IterableDiffers }, { token: i0.ChangeDetectorRef }, { token: i0.NgZone }, { token: i0.ElementRef }, { token: i0.TemplateRef }, { token: i0.ViewContainerRef }, { token: i1.RxStrategyProvider }, { token: i0.ErrorHandler }], target: i0.ɵɵFactoryTarget.Directive });
/** @nocollapse */ /** @nocollapse */ RxFor.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.0", type: RxFor, selector: "[rxFor]", inputs: { rxFor: "rxFor", rxForOf: "rxForOf", strategy: ["rxForStrategy", "strategy"], renderParent: ["rxForParent", "renderParent"], patchZone: ["rxForPatchZone", "patchZone"], trackBy: ["rxForTrackBy", "trackBy"], renderCallback: ["rxForRenderCallback", "renderCallback"] }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: RxFor, decorators: [{
            type: Directive,
            args: [{
                    selector: '[rxFor]',
                }]
        }], ctorParameters: function () { return [{ type: i0.IterableDiffers }, { type: i0.ChangeDetectorRef }, { type: i0.NgZone }, { type: i0.ElementRef }, { type: i0.TemplateRef }, { type: i0.ViewContainerRef }, { type: i1.RxStrategyProvider }, { type: i0.ErrorHandler }]; }, propDecorators: { rxFor: [{
                type: Input
            }], rxForOf: [{
                type: Input
            }], strategy: [{
                type: Input,
                args: ['rxForStrategy']
            }], renderParent: [{
                type: Input,
                args: ['rxForParent']
            }], patchZone: [{
                type: Input,
                args: ['rxForPatchZone']
            }], trackBy: [{
                type: Input,
                args: ['rxForTrackBy']
            }], renderCallback: [{
                type: Input,
                args: ['rxForRenderCallback']
            }] } });

const DECLARATIONS = [
    RxFor
];
class ForModule {
}
/** @nocollapse */ /** @nocollapse */ ForModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: ForModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
/** @nocollapse */ /** @nocollapse */ ForModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: ForModule, declarations: [RxFor], exports: [RxFor] });
/** @nocollapse */ /** @nocollapse */ ForModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: ForModule, imports: [[]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.0", ngImport: i0, type: ForModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: DECLARATIONS,
                    imports: [],
                    exports: DECLARATIONS
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { ForModule, RxFor };
//# sourceMappingURL=template-experimental-for.mjs.map
